/*  PNMNORMSCANS - normalise brightness of scans to remove banding artifacts
 *
 * This file is part of PhotoPNMTools,
 * Release $Name:  $,
 * File rev. $Id: pnmnormscans.c,v 1.2 2003/02/22 12:21:19 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>

#define MAX(a,b) ( (a) > (b) ? (a) : (b) )

#define ABS(a) ( (a) > 0 ? (a) : -(a) )

#define USAGE \
"pnmnormscans: normalise brightness of scanlines heuristically,\n" \
"reducing horizontal band artifacts found in typical photo scans\n" \
"options: -a  do analysis only, and show histogram\n" \
"         -p [banding period, integer]\n" \
"         -h [number of luminance levels in histogram, default=32]\n" \
"         -t [threshold of significance, a float, default=1.0]\n" \
"         -r [offset of reference scan line (0...period-1), default 0]\n" \
"         -l [do not modify pixels above this brightness level, default=0.5]\n" \
"         -s [strengt of smoothing step after normalisation (0..1) 0=none]\n" \
"\n" \

enum {Normal,Analysis};

	int
main(int argc, char **argv) {
	xel **in,**out;

	int mode=Normal;

	int period=3;
	int histsize=32;
	int basescanofs = 0;
	float normthreshold=0.5, norm;
	float maxbright=0.5;
	int thresh,maxpixbright;

	float smoothing=0.25;
	float rf,gf,bf, lumf;
	float weight;

	long r,g,b, lum;
	int idx;
	int basescan;
	long lumdif, predlumdif;

	char optch;
	long *hist;
	long *histnr;
	pixval maxpix;
	int format;
	int xsize,ysize;
	int x,y;
	pnm_init(&argc,argv);
	while ( (optch = getopt(argc, argv, "a")) != -1) {
		switch (optch) {
		case 'a': mode=Analysis;
		break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	in = pnm_readpnm(stdin,&xsize,&ysize,&maxpix,&format);
	out = pnm_allocarray(xsize,ysize);
	hist = (long *)calloc(histsize * period * 3,sizeof(long));
	if (!hist) exit(EXIT_FAILURE);
	histnr = (long *)calloc(histsize * period,sizeof(long));
	if (!hist) exit(EXIT_FAILURE);

	/* determine histogram of luminance differences from the basescans
	 * by the relevant criteria rgb level and scan offset */
	/* note that the histogram values for the basescans are just zeroes */
	for (y=0; y<ysize; y++) {
		basescan = ((y+1)/period)*period + basescanofs;
		if (basescan >= ysize) break;
		for (x=0; x<xsize; x++) {
			lum=((in[y][x].r+in[y][x].g+in[y][x].b) * histsize)/((maxpix+1)*3);
			idx = 3 * ( lum + histsize*(y%period));
			hist  [idx  ] += in[y][x].r - in[basescan][x].r;
			hist  [idx+1] += in[y][x].g - in[basescan][x].g;
			hist  [idx+2] += in[y][x].b - in[basescan][x].b;
			histnr[idx  ]++;
		}
	}

	norm = (float)(xsize*ysize)/(float)(histsize*3);

	for (x=0; x<histsize; x++) {
		for (y=0; y<period; y++) {
			fprintf(stderr,"%f\t",
					(float)hist[3 * (x + histsize*y)] / (float)norm);
		}
		fprintf(stderr,"\n");
	}

	thresh = (xsize*ysize*normthreshold)/(histsize*3);
	maxpixbright = 3*maxpix*maxbright;
	/* correct rgb levels according to histogram */
	for (y=0; y<ysize; y++) {
		basescan = ((y+1)/period)*period + basescanofs;
		if (basescan >= ysize) break;
		for (x=0; x<xsize; x++) {
			lum=((in[y][x].r+in[y][x].g+in[y][x].b) * histsize)/((maxpix+1)*3);
			idx = 3 * ( lum + histsize*(y%period));
			r = in[y][x].r;
			g = in[y][x].g;
			b = in[y][x].b;
			if (histnr[idx] != 0
			&& ABS(hist[idx]) > thresh
			&& r+g+b < maxpixbright) {
				r -= (hist[idx  ] / histnr[idx]);
				g -= (hist[idx+1] / histnr[idx]);
				b -= (hist[idx+2] / histnr[idx]);
			}
			if (r < 0) r=0;
			if (r > maxpix) r=maxpix;
			if (g < 0) g=0;
			if (g > maxpix) g=maxpix;
			if (b < 0) b=0;
			if (b > maxpix) b=maxpix;
			out[y][x].r=r;
			out[y][x].g=g;
			out[y][x].b=b;
		}
	}

	/* optionally, do some smoothing to get rid of banding completely */
	/* copy out back to in */
	for (y=0; y<ysize; y++) {
		for (x=0; x<xsize; x++) {
			in[y][x].r=out[y][x].r;
			in[y][x].g=out[y][x].g;
			in[y][x].b=out[y][x].b;
		}
	}
	for (y=0; y<ysize; y++) {
		for (x=0; x<xsize; x++) {
			rf = in[y][x].r;
			gf = in[y][x].g;
			bf = in[y][x].b;
			lumf = (rf + bf + gf) / 3.0 / maxpix;
			weight = 1;
			if (y > 0 && lumf < maxbright) {
				rf += smoothing*out[y-1][x].r;
				gf += smoothing*out[y-1][x].g;
				bf += smoothing*out[y-1][x].b;
				weight += smoothing;
			}
			if (y < ysize - 1 && lumf < maxbright) {
				rf += smoothing*out[y+1][x].r;
				gf += smoothing*out[y+1][x].g;
				bf += smoothing*out[y+1][x].b;
				weight += smoothing;
			}
			out[y][x].r = rf / weight;
			out[y][x].g = gf / weight;
			out[y][x].b = bf / weight;
		}
	}
	pnm_writepnm(stdout,out,xsize,ysize,maxpix,format,0/*forceplain*/);
	pnm_freearray(in,ysize);
	pnm_freearray(out,ysize);
	exit(EXIT_SUCCESS);
	return 0;
}

