/*  PNMREMFRINGES - remove specific fringes from a specific scanner :-)
 *
 * This file is part of PhotoPNMTools,
 * Release $Name:  $,
 * File rev. $Id: pnmremfringes.c,v 1.2 2003/02/22 12:21:19 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>

#define MAX(a,b) ( (a) > (b) ? (a) : (b) )
#define MIN(a,b) ( (a) < (b) ? (a) : (b) )

#define MAX3(a,b,c) ( (a) > (b) ? MAX((a),(c)) : MAX((b),(c)) )
#define MIN3(a,b,c) ( (a) < (b) ? MIN((a),(c)) : MIN((b),(c)) )

#define ABS(a) ( (a) > 0 ? (a) : -(a) )

#define USAGE \
"pnmremfringes: remove colour fringing artefacts from scans\n" \
"\n" \

enum {Normal,Analysis};

// routine adapted from Nan C. Schaller
// red = 0, yellow = 60, green = 120, cyan = 180, blue = 240, magenta = 300
// r,g,b values are from 0 to 1
// h = [0,360], s = [0,1], v = [0,1]
//              if s == 0, then h = -1 (undefined)
void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v ) {
	float min, max, delta;

	min = MIN3( r, g, b );
	max = MAX3( r, g, b );
	*v = max;			       // v

	delta = max - min;

	if( max != 0 )
		*s = delta / max;	       // s
	else {
		// r = g = b = 0		// s = 0, v is undefined
		*s = 0;
		*h = -1;
		return;
	}

	if( r == max )
		*h = ( g - b ) / delta;	 // between yellow & magenta
	else if( g == max )
		*h = 2 + ( b - r ) / delta;     // between cyan & yellow
	else
		*h = 4 + ( r - g ) / delta;     // between magenta & cyan

	*h *= 60;			       // degrees
	if( *h < 0 )
		*h += 360;
}

// routine adapted from Nan C. Schaller
void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v ) {
	int i;
	float f, p, q, t;

	if( s == 0 ) {
		// achromatic (grey)
		*r = *g = *b = v;
		return;
	}

	h /= 60;			// sector 0 to 5
	i = floor( h );
	f = h - i;		      // factorial part of h
	p = v * ( 1 - s );
	q = v * ( 1 - s * f );
	t = v * ( 1 - s * ( 1 - f ) );

	switch( i ) {
		case 0:
			*r = v;
			*g = t;
			*b = p;
			break;
		case 1:
			*r = q;
			*g = v;
			*b = p;
			break;
		case 2:
			*r = p;
			*g = v;
			*b = t;
			break;
		case 3:
			*r = p;
			*g = q;
			*b = v;
			break;
		case 4:
			*r = t;
			*g = p;
			*b = v;
			break;
		default:		// case 5:
			*r = v;
			*g = p;
			*b = q;
			break;
	}
}

	float
lum_diff(xel *xel1, xel *xel2, long maxpix) {
	/* returns a value between 0 and 1 */
	float lum1 = (xel1->r + xel1->g + xel1->b) / (float)(maxpix*3.0);
	float lum2 = (xel2->r + xel2->g + xel2->b) / (float)(maxpix*3.0);
	return ABS(lum1-lum2);
}

	float
red_diff(xel *xel1, xel *xel2, long maxpix) {
	/* returns a value between 0 and 1 */
	float lum1 = xel1->r / (float)(maxpix);
	float lum2 = xel2->r / (float)(maxpix);
	return ABS(lum1-lum2);
}

	float
green_diff(xel *xel1, xel *xel2, long maxpix) {
	/* returns a value between 0 and 1 */
	float lum1 = xel1->g / (float)(maxpix);
	float lum2 = xel2->g / (float)(maxpix);
	return ABS(lum1-lum2);
}

	float
hue_diff(float hue1, float hue2) {
	/* number between 0 (same) and 1 (180 different) */
	int diff = ABS(hue1 - hue2);
	if (diff > 180) diff = 360 - diff;
	return (float)diff / 180.0;
}

	int
main(int argc, char **argv) {
	xel **in,**out;

	int mode=Normal;
	long floor,ceil;
	long r,g,b;
	char optch;

	pixval maxpix;
	int format;
	int xsize,ysize;
	int x,y,dx;

	int modified;

	float rf,bf,gf;
	float h,s,v;
	float lh,ls,lv;
	float rh,rs,rv;
	
	float contrast, rcontrast, gcontrast, bcontrast, hue;
	pnm_init(&argc,argv);
	while ( (optch = getopt(argc, argv, "esdf:c:")) != -1) {
		switch (optch) {
		case 'a': mode=Analysis;
		break;
		case 'l': floor=atol(optarg);
		break;
		case 'u': ceil=atol(optarg);
		break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	in = pnm_readpnm(stdin,&xsize,&ysize,&maxpix,&format);
	out = pnm_allocarray(xsize,ysize);

	/* find suspicious reds and cyans within contrasty areas */
	/* red: more or less pure red */
	/* cyan: mostly green with some blue */
	for (y=0; y<ysize; y++) {
		for (x=0; x<xsize; x++) {
			out[y][x] = in[y][x];
		}
	}
	for (y=8; y<ysize-8; y++) {
		for (x=8; x<xsize-8; x++) {
			rcontrast=0;
			gcontrast=0;
			rf = (float)in[y][x].r / (float)maxpix;
			gf = (float)in[y][x].g / (float)maxpix;
			bf = (float)in[y][x].b / (float)maxpix;
			for (dx=1; dx<8; dx++) {
				rcontrast +=
					(float)red_diff(&(in[y][x+dx]),&(in[y][x-dx]),maxpix)
					* (1.0/dx);
				gcontrast +=
					(float)green_diff(&(in[y][x+dx]),&(in[y][x-dx]),maxpix)
					* (1.0/dx);
			}
			/* find hsv of pixels to the left and right of this pixel */
			RGBtoHSV(in[y][x-8].r, in[y][x-8].g, in[y][x-8].b, &lh,&ls,&lv);
			RGBtoHSV(in[y][x+8].r, in[y][x+8].g, in[y][x+8].b, &rh,&rs,&rv);
			/* hsv of this pixel */
			RGBtoHSV(rf,gf,bf, &h,&s,&v);
			if (s > 0.15) {
				modified=0;
				/* 300...10 */
				if( (  (h > 290 || h < 20) && rcontrast > 0.2 /* red */
						&& (hue_diff(h,lh) > 0.15 || ls < 0.15)
						&& (hue_diff(h,rh) > 0.15 || rs < 0.15) ) ) {
					s /= 1.0 + 10.0 * (rcontrast-0.2);
					modified=1;
				}
				/* 90...150 */
				if( ( (h > 80 && h <160 && gcontrast > 0.2) /* cyan */
					 	&& ( hue_diff(h,lh) > 0.15 || ls < 0.15)
						&& ( hue_diff(h,rh) > 0.15 || rs < 0.15) ) ) {
					s /= 1.0 + 10.0 * (gcontrast-0.2);
					modified=1;
				}
				if (modified) {
					HSVtoRGB(&rf,&gf,&bf, h,s,v);
					out[y][x].r = rf * maxpix;
					out[y][x].g = gf * maxpix;
					out[y][x].b = bf * maxpix;
				}
			}
		}
	}

	pnm_writepnm(stdout,out,xsize,ysize,maxpix,format,0/*forceplain*/);
	pnm_freearray(in,ysize);
	pnm_freearray(out,ysize);
	exit(EXIT_SUCCESS);
	return 0;
}

