/*  PPMHSY - improved hsv-like colour model
 *
 * This file is part of PhotoPNMTools,
 * Release $Name:  $,
 * File rev. $Id: ppmhsy.c,v 1.5 2003/03/08 17:17:21 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

/* Alternative models:
 * cieluv cielhv? cielhs? and tekhvc
 * http://www.powerretouche.com/SaturationUserguide.htm
 * hsv (=hsb?)i=max(r,g,b); 
 * hsi i=r+g+b
 * hsb
 * hls i=max+min
 * hwb (hue-whiteness-blackness) whiteness=min(r,g,b) blackness=max(r,g,b)
 * hvc?
 * lhs: http://www.csis.hku.hk/~yang/color/ch2.ps
 *
 * hls: http://www.aboutvb.de/khw/artikel/khwrgbhlshsv.htm
 * nMaxMinDelta = nMax - nMin
   If rL <= HLSmax / 2 Then
     rS = ((nMaxMinDelta * HLSmax) + ((nMax + nMin) / 2)) / _
      (nMax + nMin)
   Else
     rS = ((nMaxMinDelta * HLSmax) + _
      ((2 * kRGBmax - nMax - nMin) / 2)) / _
      (2 * kRGBmax - nMax - nMin)
   End If
 * 
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>

#include "stdpnm.h"

#define USAGE \
"ppmhsy: do hue-saturation-brightness operations on image.\n" \
"    V1.7 4 feb 2003, Boris van Schooten   boris@13thmonkey.org\n" \
"Uses improved brightness definition Y = R*Rweight + G*Gweight + B*Bweight\n" \
"to retain perceptual brightness when changing saturation or hue.  \n" \
"Changes that cause r, g, or b to fall outside their maxvalues can be \n" \
"compressed by adding extra white, or the image can be darkened to allow\n" \
"enough room for them.\n" \
"    -I : input is HSY (and default output is 8-bit RGB)\n" \
"    -O : output is 16-bit HSY\n" \
"    -s [saturation mul factor, default 1.0]    : change saturation\n" \
"    -0 .. -5 [hue-specific saturation mul factor, default 1.0]\n" \
"       : change saturation, 0=red 1=yellow 2=green 3=cyan 4=blue 5=magenta\n"\
"    -h [hue-shift, -360.0..360.0, default 0.0] : change hue\n" \
"    -y [brightness-mul-factor, default 1.0]    : change brightness\n" \
"    -Y [gamma-factor, default 1.0]             : change gamma\n" \
"    -c [contrast-mul-factor, >=0, default 1.0] : change contrast\n" \
"    -w [whiten-factor, 0.0..1.0, default 1.0]\n" \
"                   : amount to whiten if r, g, or b out of range\n" \
"                     (0.0 = don't whiten, 1.0 = try to retain brightness)\n" \
"    -d [darken-factor, 0.0..1.0, default 1.0]\n" \
"                   : amount to darken to fit rgb in (0.0=black, 1.0=none)\n" \
"    -n [saturation noise dampening threshold, 0.0..1.0, default 0.0]\n" \
"                   : brightness below which saturation is decreased\n" \
"    -E             : Use exponentional formula for colour balance\n" \
"    -R [brightness of red,   default 0.299]\n" \
"    -G [brightness of green, default 0.578]\n" \
"    -B [brightness of blue,  default 0.114]\n" \
"    -r [red-mul-factor, >=0.0, default 1.0]    : red-cyan colour balance\n" \
"    -g [green-mul-factor, >=0.0, default 1.0]  : green-magenta colour balance\n" \
"    -b [blue-mul-factor, >=0.0, default 1.0]   : blue-yellow colour balance\n"


	double
mid3(double a,double b,double c) {
	double min = MIN3(a,b,c);
	double max = MAX3(a,b,c);
	if (a==min) {
		if (b==max) return c;
		else        return b;
	} else if (b==min) {
		if (a==max) return c;
		else        return a;
	}
	/* c==min */
	if (a==max) return b;
	else        return a;
}

/* sum must be 3 */
double rweight=0.299,gweight=0.587,bweight=0.114;
/* effective: 0.8, 1.5, 0.6 */
/* gimp: 150,76,28 */
/* thesis: 0.299,0.587,0.114 */

double red_shift=1.0,green_shift=1.0,blue_shift=1.0;

double whitenfactor=1.0;
/* whiten if r, g, or b out of range to retain original brightness */

double darkenfactor=1.0;
/* darken if rgb out of range. Range: [0.0-1.0]. 1.0 means no darkening. */

double sat_dampening_thresh=0.0;

/* maximum brightness stored in hsv-format ppm file */
double maxbright=2.0;

double Gamma = 1.0;

int exponential=0;

struct hsv {
	double h,s,v;
};

double satmulh[6];

	struct hsv
xeltohsv(xel hsvxel, int maxpix) {
	struct hsv hsv;
	hsv.h = (double)hsvxel.r / (double)maxpix * 359.999999;
	hsv.s = (double)hsvxel.g / (double)maxpix;
	hsv.v = (double)hsvxel.b / (double)maxpix * maxbright;
	return hsv;
}

	xel
hsvtoxel(struct hsv hsv, int maxpix) {
/*** NOTE: clips values <0.0 or >1.0 */
	xel hsvxel;
	double v=hsv.v;
	double s=hsv.s;
	if (hsv.h < 0.0 || hsv.h >= 360.0)
		fprintf(stderr,"hsvtoxel: unexpected h %f",hsv.h);
	if (s > 1.0) {/*fprintf(stderr,"r:%f\n",r);*/ s = 1.0; }
	if (s < 0.0) {/*fprintf(stderr,"r:%f\n",r);*/ s = 0.0; }
	if (v > maxbright) {/*fprintf(stderr,"r:%f\n",r);*/ v = maxbright; }
	if (v < 0.0)       {/*fprintf(stderr,"b:%f\n",b);*/ v = 0.0; }
	hsvxel.r = hsv.h / 360.0 * maxpix;
	hsvxel.g =     s * maxpix;
	hsvxel.b =     v / maxbright * maxpix;
	return hsvxel;
}

	struct hsv
rgbtohsv(struct rgb rgb) {
	/* some useful intermediate variables */
	double r = rgb.r;
	double g = rgb.g;
	double b = rgb.b;
	double rw = r * rweight;
	double gw = g * gweight;
	double bw = b * bweight;
	double max = MAX3(r,g,b);
	double min = MIN3(r,g,b);
	double mid = mid3(r,g,b);
	double maxw,midw; // weight of max and mid
	double delta = max - min;

	struct hsv hsv;
	/* special case for black */
	if (max == 0.0) {
		hsv.h = 0; /* note h is actually undefined */
		hsv.s = 0;
		hsv.v = 0;
		return hsv;
	}
	/* get hue */
	if (max==mid && mid==min) /* grey */
		hsv.h = 0; /* h undefined */
	else if( r == max )
		hsv.h = ( g - b ) / delta;  // between yellow & magenta
	else if( g == max )
		hsv.h = 2 + ( b - r ) / delta;     // between cyan & yellow
	else
		hsv.h = 4 + ( r - g ) / delta;     // between magenta & cyan

	hsv.h *= 60.0;                              // degrees
	if( hsv.h < 0.0 )
		hsv.h += 360.0;
	/* get weights in a way that is exactly symmetrical to hsvtorgb to ensure
	 * the cases where max=mid or mid=min are handled in the same way */
	switch ( (int)(floor(hsv.h / 60.0000001)) ) {
		case 0 : /* red..yellow */
			maxw = rweight;
			midw = gweight;
		break;
		case 1 : /* yellow..green*/
			maxw = gweight;
			midw = rweight;
		break;
		case 2 : /* green..cyan */
			maxw = gweight;
			midw = bweight;
		break;
		case 3 : /* cyan..blue */
			maxw = bweight;
			midw = gweight;
		break;
		case 4 : /* blue..magenta */
			maxw = bweight;
			midw = rweight;
		break;
		case 5 : /* magenta..red */
			maxw = rweight;
			midw = bweight;
		break;
		default:
			fprintf(stderr,"rgbtohsv: Unexpected h %f\n",hsv.h);
			exit(EXIT_FAILURE);
		break;
	}
	hsv.v = rw + gw + bw;
	hsv.s = (maxw*(max-min) + midw*(mid-min)) / hsv.v;
	hsv.v /= 3.0;
	/*if (max==r) {
		maxw=rweight;
		if (mid==g) midw=gweight;
		else        midw=bweight;
	} else if (max==g) {
		maxw=gweight;
		if (mid==r) midw=rweight;
		else        midw=bweight;
	} else {
		maxw=bweight;
		if (mid==r) midw=rweight;
		else        midw=gweight;
	}*/
	return hsv;
}


	struct rgb
hsvtorgb(struct hsv hsv) {
	double r,g,b;
	double hsect,hfrac;
	double max,mid,min, newmin,newmid,newmax, new2min,new2mid,new2max;
	double maxw,midw,minw;
	double huescale;
	double lostbright;
	//double rov,gov,bov, totov;
	//int nrov;
	struct rgb rgb;
	/* special case for grey */
	if (hsv.s == 0.0) {
		rgb.r = hsv.v * darkenfactor;
		rgb.g = hsv.v * darkenfactor;
		rgb.b = hsv.v * darkenfactor;
        return rgb;
	}
	if (hsv.h < 0.0 || hsv.h >= 360.0)
		fprintf(stderr,"hsvtorgb: Unexpected h %f\n",hsv.h);
	//if (hsv.s < 0) {
	//	fprintf(stderr,"asdfseafasdfdsafdfas 2\n");
	//	exit(EXIT_FAILURE);
	//}
	hfrac = hsv.h / 60.0;
	hsect = floor( hfrac ); // sector 0 to 5
	hfrac = hfrac - hsect; // factorial part of h
	hsv.v *= 3.0;
	/* determine m..w */
	switch ((int)(hsect)) {
		case 0 : /* red..yellow */
			maxw = rweight;
			midw = gweight;
			minw = bweight;
		break;
		case 1 : /* yellow..green*/
			maxw = gweight;
			midw = rweight;
			minw = bweight;
		break;
		case 2 : /* green..cyan */
			maxw = gweight;
			midw = bweight;
			minw = rweight;
		break;
		case 3 : /* cyan..blue */
			maxw = bweight;
			midw = gweight;
			minw = rweight;
		break;
		case 4 : /* blue..magenta */
			maxw = bweight;
			midw = rweight;
			minw = gweight;
		break;
		case 5 : /* magenta..red */
			maxw = rweight;
			midw = bweight;
			minw = gweight;
		break;
		default:
			fprintf(stderr,"hsvtorgb 2: Unexpected h %f\n",hsv.h);
			exit(EXIT_FAILURE);
		break;
	}

/*	min = (hsv.s * hsv.v - hsv.v) / -3.0;
	//fprintf(stderr,"min:%f ",min);
	if ((int)hsect % 2 == 1) {
		hfrac -= 1.0;
		mid = (hfrac*hsv.v - 2.0*hfrac*min - min) / (hfrac - 1.0);
		//fprintf(stderr,"mid:%f \n",mid);
	} else {
		mid = (hfrac*hsv.v - 2.0*hfrac*min + min) / (hfrac + 1.0);
	}
	max = hsv.v - mid - min; */

	min = (hsv.s * hsv.v - hsv.v) / -3.0;
	//fprintf(stderr,"min:%f ",min);
	if ((int)hsect % 2 == 1) {
		hfrac -= 1.0;
		mid = (  min + hfrac*( min + (minw*min - hsv.v)/maxw )  )
			 /(  1 - hfrac*(midw/maxw)  );
		//fprintf(stderr,"mid:%f \n",mid);
	} else {
		mid = (  min + hfrac * ( (hsv.v - minw*min)/maxw - min )  )
			 /(  1 + hfrac*(midw/maxw)  );
	}
	max = ( hsv.v - midw*mid - minw*min ) / maxw;
	/* darken */
	max *= darkenfactor;
	mid *= darkenfactor;
	min *= darkenfactor;
	/* deal with overflows in case s > 1.0 or v > 1.0 */
	if (max > 1.0 && min < 1.0) {
		huescale = (mid-min)/(max-min);
		/* shrink to make hue fit in remaining bit of rgb space */
		newmid = min + (1.0-min) * huescale;
		newmax = 1.0;
		newmin = min;
		/* whiten to add lost brightness */
		lostbright = whitenfactor*((max - newmax)*maxw + (mid - newmid)*midw);
		new2min = (  minw*newmin + lostbright - midw*(huescale - newmid)  )/
			     (  midw*(1 - huescale) + minw  );
		new2mid = new2min + (1.0-new2min) * huescale;
		new2max = newmax;
		min = new2min;
		mid = new2mid;
		max = new2max;
		/*newmin = min + whitenfactor*(max-1.0)/minw; */
	} else if (min >= 1.0) {
			min = 1.0;
			mid = 1.0;
			max = 1.0;
	}
	switch ((int)(hsect)) {
		case 0 : /* red..yellow */
			r = max;
			g = mid;
			b = min;
		break;
		case 1 : /* yellow..green*/
			g = max;
			r = mid;
			b = min;
		break;
		case 2 : /* green..cyan */
			g = max;
			b = mid;
			r = min;
		break;
		case 3 : /* cyan..blue */
			b = max;
			g = mid;
			r = min;
		break;
		case 4 : /* blue..magenta */
			b = max;
			r = mid;
			g = min;
		break;
		case 5 : /* magenta..red */
			r = max;
			b = mid;
			g = min;
		break;
		default:
			fprintf(stderr,"hsvtorgb: Unexpected h %f\n",hsv.h);
			exit(EXIT_FAILURE);
		break;
	}
	/*
	rov = r-1.0;
	gov = g-1.0;
	bov = b-1.0;
	nrov = 0;
	totov = 0.0;
	if (rov > 0.0) { nrov++; totov += rov; }
	if (gov > 0.0) { nrov++; totov += gov; }
	if (bov > 0.0) { nrov++; totov += bov; }
	if (nrov == 3) exit(1);
	if (nrov) {
		if (rov < 0.0) r += totov/2.0;
		if (gov < 0.0) g += totov/2.0;
		if (bov < 0.0) b += totov/2.0;
	}
	*/
	rgb.r = r;
	rgb.g = g;
	rgb.b = b;
	return rgb;
}


	int
main(int argc, char **argv) {
	xel **in,**out;
	//xel *pixel;
	//struct hsv *hsv;

	/* settings */
	double satmul=1.0;
	double valmul=1.0;
	double hueshift=0.0;
	double contrast=1.0;
	int hsvin=0,hsvout=0;
	
	double rgbtot;
	double h, hdist;
	//long r,g,b;
	char optch;

	struct hsv hsv,hsv2;
	struct rgb rgb;
	pixval srcmaxpix,dstmaxpix;
	int format;
	int xsize,ysize;
	int x,y, i;
	double maxshift;

	for (i=0; i<6; i++) satmulh[i]=1.0;
	pnm_init(&argc,argv);
	while ( (optch = getopt(argc, argv,
	"IOs:y:Y:h:c:w:d:n:ER:G:B:r:g:b:0:1:2:3:4:5:")) != -1) {
		switch (optch) {
		case 'I': hsvin = 1;
		break;
		case 'O': hsvout = 1;
		break;
		case 's': satmul=atof(optarg);
		break;
		case 'y': valmul=atof(optarg);
		break;
		case 'Y': Gamma=atof(optarg);
		break;
		case 'h': hueshift=atof(optarg);
		break;
		case 'c': contrast=atof(optarg);
		break;
		case 'w': whitenfactor=atof(optarg);
		break;
		case 'd': darkenfactor=atof(optarg);
		break;
		case 'n': sat_dampening_thresh=atof(optarg);
		break;
		case 'E': exponential = 1;
		break;
		case 'R': rweight=atof(optarg);
		break;
		case 'G': gweight=atof(optarg);
		break;
		case 'B': bweight=atof(optarg);
		break;
		case 'r': red_shift=atof(optarg);
		break;
		case 'g': green_shift=atof(optarg);
		break;
		case 'b': blue_shift=atof(optarg);
		break;
		case '0': satmulh[0]=atof(optarg);
		break;
		case '1': satmulh[1]=atof(optarg);
		break;
		case '2': satmulh[2]=atof(optarg);
		break;
		case '3': satmulh[3]=atof(optarg);
		break;
		case '4': satmulh[4]=atof(optarg);
		break;
		case '5': satmulh[5]=atof(optarg);
		break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	/* ensure r+g+b = 3 */
	rgbtot = rweight + gweight + bweight;
	rweight *= 3.0 / rgbtot;
	gweight *= 3.0 / rgbtot;
	bweight *= 3.0 / rgbtot;
	//fprintf(stderr,"r%f g%f b%f\n",rweight,gweight,bweight);
	in = pnm_readpnm(stdin,&xsize,&ysize,&srcmaxpix,&format);
	out = pnm_allocarray(xsize,ysize);
	//hsv = malloc(sizeof(struct hsv)*xsize*ysize);
	//if (!hsv) exit(EXIT_FAILURE);

	/* ensure colour shifts are within range 0..1 */
	maxshift = MAX3(red_shift,green_shift,blue_shift);
	if (maxshift > 1.0) {
		red_shift   /= maxshift;
		green_shift /= maxshift;
		blue_shift  /= maxshift;
	}
	
	dstmaxpix = srcmaxpix;
	if (hsvin) {
		dstmaxpix = 255; // 8 bit RGB
	}
	if (hsvout) {
		dstmaxpix = 65535; // hsv output is always 16 bit
	}

	for (y=0; y<ysize; y++) {
		for (x=0; x<xsize; x++) {
			if (hsvin) {
				hsv = xeltohsv(in[y][x],srcmaxpix);
				rgb = hsvtorgb(hsv);
			} else {
				rgb = xeltorgb(in[y][x],srcmaxpix);
				hsv = rgbtohsv(rgb);
			}
			if (sat_dampening_thresh > 0.0) {
				// s decreases linearly to 0
				if (hsv.v < sat_dampening_thresh) {
					hsv.s *= 1.0 - 
					              (1.0/sat_dampening_thresh)
						        * (sat_dampening_thresh - hsv.v);
					rgb = hsvtorgb(hsv);
				}
			}
			if (red_shift!=1.0 || green_shift!=1.0 || blue_shift!=1.0) {
				//shiftmax = MAX3(red_shift,green_shift,blue_shift);
				if (!exponential) {
					rgb.r *= red_shift;
					rgb.g *= green_shift;
					rgb.b *= blue_shift;
				} else {
					/* alternative: shifts are gamma values */
					rgb.r = pow(rgb.r, red_shift);
					rgb.g = pow(rgb.g, green_shift);
					rgb.b = pow(rgb.b, blue_shift);
				}
				hsv2 = rgbtohsv(rgb);
				// preserve luminance by inserting old brightness.
				hsv.h = hsv2.h;
				hsv.s = hsv2.s;
			}
			//fprintf(stderr,"h%f s%f v%f\n",hsv.h,hsv.s,hsv.v);
			hsv.s *= satmul;
			if (hsv.s > 1.0) hsv.s=1.0;
			for (i=0; i<6; i++) {
				if (satmulh[i]!=1.0) {
					h = hsv.h;
					if (i<2 && h > 180.0) h -= 360.0;
					if (i>4 && h < 180.0) h += 360.0;
					hdist = fabs(h - 60.0*i) / 30.0;
					if (hdist < 1.0) {
						hsv.s *= satmulh[i];
					} else if (hdist < 1.5) {
						hdist = (hdist - 1.0)/0.5;
						hsv.s *= satmulh[i]*(1.0 - hdist) + hdist;
					}
				}
			}
			hsv.v *= valmul;
			if (contrast != 1.0) {
				// we do contrast operations in hsy instead of rgb, which
				// gives a much more natural result.  However, decreasing
				// contrast amplifies saturation noise in the dark colours.
				hsv.v = (hsv.v - 0.5)*contrast + 0.5;
				// we do not deal with brightness < 0
			}
			if (hsv.v < 0.0) hsv.v=0.0;
			if (Gamma != 1.0) {
				hsv.v = pow(hsv.v,Gamma);
			}
			//if (hsv.v > 1.0) hsv.v=1.0;
			hsv.h += hueshift;
			if (hsv.h < 0.0) hsv.h += 360.0;
			if (hsv.h >= 360.0) hsv.h -= 359.999999;
			if (hsvout) {
				out[y][x] = hsvtoxel(hsv,dstmaxpix);
			} else {
				rgb = hsvtorgb(hsv);
				out[y][x] = rgbtoxel(rgb,dstmaxpix);
			}
		}
	}

	pnm_writepnm(stdout,out,xsize,ysize,dstmaxpix,format,0/*forceplain*/);
	pnm_freearray(in,ysize);
	pnm_freearray(out,ysize);
	exit(EXIT_SUCCESS);
	return 0;
}

